<?php
// $Id: operations.pipeline.inc,v 1.2 2009/03/24 19:44:59 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2008 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */

/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfPipelineOperation() {
  return array(
    'label' => t('Pipeline'),
    'description' => t('Executes a whole pipeline as a single operation.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_operation_TfPipelineOperation_type_options(&$typeOptionsInfo, $prefix) {
  $typeOptionsInfo[$prefix . 'pipeline']['label'] = t('Pipeline');

  $query = 'SELECT pipeline_id, name FROM {transformations_pipeline}';
  $params = array();

  if (isset($typeOptionsInfo[$prefix . 'pipeline']['value'])) {
    $query .= ' WHERE pipeline_id = %d';
    $params[] = $typeOptionsInfo[$prefix . 'pipeline']['value'];
  }
  $result = db_query($query, $params);
  $pipelines = array();

  while ($pipeline = db_fetch_object($result)) {
    $pipelines[$pipeline->pipeline_id] = check_plain($pipeline->name);
  }
  $typeOptionsInfo[$prefix . 'pipeline']['options'] = $pipelines;
}


/**
 * An operation to execute a TfPipeline. Multiple TfPipelineOperations can run
 * off the same TfPipeline object, but that object is expected not to change
 * while this operation is still in use (otherwise, bugs are likely to appear).
 */
class TfPipelineOperation extends TfOperation implements TfOutputSchemaListener {
  private $pipeline;

  /** An array of actual operation instances, one per operationId
   * (which is the array key for each element).
   * NULL if the operations have not yet been instantiated. */
  private $operations;

  /** Cached result of the hasSideEffects() method. */
  private $hasSideEffects;

  /** If pipeline execution failed, this contains the operation identifier
   * of the operation that did not deliver a valid output. */
  private $failedOperationId;


  /**
   * Implementation of TfOperation::initialize().
   */
  public function initialize($typeOptions = array()) {
    if (isset($this->operations)) {
      foreach ($this->operations as $operationId => $operation) {
        $operation->unregisterOutputSchemaListener($this);
      }
    }
    parent::initialize($typeOptions);

    $this->pipeline = FALSE;
    $this->operations = NULL;
    $this->hasSideEffects = NULL;
    $this->failedOperationId = NULL;

    if (is_string($typeOptions['pipeline']) || is_numeric($typeOptions['pipeline'])) {
      $this->pipeline = TfPipeline::load($typeOptions['pipeline']);
    }
    elseif (is_array($typeOptions['pipeline'])) {
      $this->pipeline = TfPipeline::import($typeOptions['pipeline']);
    }
    elseif (is_object($typeOptions['pipeline']) && $typeOptions['pipeline'] instanceof TfPipeline) {
      $this->pipeline = $typeOptions['pipeline'];
    }

    if (!is_object($this->pipeline)) {
      $name_or_id = is_array($typeOptions['pipeline'])
        ? check_plain($typeOptions['pipeline']['name'])
        : check_plain($typeOptions['pipeline']);
      throw new Exception('Pipeline "' . $name_or_id . '" cannot be loaded!');
    }
  }

  /**
   * If pipeline execution failed, this method returns the operation identifier
   * of the operation that did not deliver a valid output. (For successful
   * pipeline executions, FALSE is returned.)
   */
  public function failedOperationId() {
    return isset($this->failedOperationId) ? $this->failedOperationId : FALSE;
  }

  /**
   * Overriding TfOperation::hasSideEffects().
   */
  public function hasSideEffects() {
    if (isset($this->hasSideEffects)) {
      return $this->hasSideEffects;
    }
    $this->ensureOperationInstances();
    $this->hasSideEffects = FALSE;

    foreach ($this->operations as $operationId => $operation) {
      if ($operation->hasSideEffects()) {
        $this->hasSideEffects = TRUE;
        break;
      }
    }
    return $this->hasSideEffects;
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    $id = $this->pipeline->id();
    return array('pipeline' => empty($id) ? $this->pipeline->export() : $id);
  }

  /**
   * Pipelines are potentially very complex, so let's cache the output.
   */
  protected function isOutputCachingEnabled() {
    return TRUE;
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array_keys($this->pipeline->targets(TfPipeline::Parameter));
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    $this->ensureOperationInstances();
    $parameterKey = $inputKey;
    $targets = $this->pipeline->targets(TfPipeline::Parameter, $parameterKey);

    if (empty($targets)) {
      throw new Exception('Invalid parameter key: ' . $parameterKey);
    }
    foreach ($targets as $targetOperationId => $targetDataSpecs) {
      foreach ($targetDataSpecs as $targetInputKey => $true) {
        $targetOperation = $this->operations[$targetOperationId];
        $targetKey = $targetInputKey;
        break; // A pipeline parameter always corresponds to a single target.
      }
    }
    return $targetOperation->inputProperty($targetKey, $propertyKey);
  }

  /**
   * Overriding TfOperation::inputSchemaChanged() in order to relay the new
   * schema to the inner operations and at the same time narrow down output
   * schema change notifications (which are handled by outputSchemaChanged()).
   */
  protected function inputSchemaChanged($inputKey) {
    $targets = $this->pipeline->targets(TfPipeline::Parameter, $inputKey);
    $schema = $this->inputProperty($inputKey, 'assignedSchema');

    foreach ($targets as $targetOperationId => $targetDataSpecs) {
      foreach ($targetDataSpecs as $targetKey => $true) {
        $this->operations[$targetOperationId]->setInputSchema($targetKey, $schema);
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array_keys($this->pipeline->sources(TfPipeline::Output));
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    $this->ensureOperationInstances();
    $source = $this->pipeline->source(TfPipeline::Output, $outputKey);

    if (empty($source)) {
      throw new Exception('Invalid output key: ' . $outputKey);
    }
    return $this->operations[$source['entity']]->outputProperty($source['key'], $propertyKey);
  }

  /**
   * Implementation of TfOutputSchemaListener::outputSchemaChanged().
   */
  public function outputSchemaChanged(TfOperation $changedOperation, $outputKey = NULL) {
    foreach ($this->operations as $operationId => $operation) {
      if ($changedOperation === $operation) {
        $targets = $this->pipeline->targets($operationId, $outputKey);
        $schema = $operation->outputProperty($outputKey, 'expectedSchema');

        foreach ($targets as $targetOperationId => $targetDataSpecs) {
          foreach ($targetDataSpecs as $targetKey => $true) {
            if ($targetOperationId == TfPipeline::Output) {
              $this->setOutputSchemaDirty($targetKey);
            }
            else { // it's an operation
              $this->operations[$targetOperationId]->setInputSchema($targetKey, $schema);
            }
          }
        }
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    if (!$this->pipeline->startOperation()) {
      throw new Exception('No start operation was set for the pipeline ' .
        check_plain($this->pipeline->name()) . '!');
    }

    // Set up our execution environment.
    $pendingOperations = array($this->pipeline->startOperation());
    $this->ensureOperationInstances();

    foreach ($this->pipeline->targets(TfPipeline::Parameter) as $parameterKey => $targets) {
      foreach ($targets as $targetOperationId => $targetDataSpecs) {
        foreach ($targetDataSpecs as $targetKey => $true) {
          $this->operations[$targetOperationId]->setInput($targetKey, $this->input($parameterKey));
        }
      }
    }

    // Continue to run until all (reachable) operations have been executed.
    while (!empty($pendingOperations)) {
      $currentOperationId = array_pop($pendingOperations);
      $currentOperationScheduled = FALSE;

      // Get the object for the currently processed operation.
      if (!isset($this->operations[$currentOperationId])) {
        throw new Exception('No operation exists with operation identifier ' .
          $currentOperationId . '.');
      }
      $currentOperation = $this->operations[$currentOperationId];

      // If the operation output is not relevant, we might just as well
      // skip executing the operation in the first place.
      if (!$currentOperation->hasSideEffects()) {
        $targets = $this->pipeline->targets($currentOperationId);
        if (empty($targets)) {
          continue; // neither side effects nor targets: skip
        }
      }

      foreach ($currentOperation->inputKeys() as $inputKey) {
        if (!$currentOperation->isInputSet($inputKey)) {
          $source = $this->pipeline->source($currentOperationId, $inputKey);

          if (!empty($source) && TfPipeline::sourceType($source) == TfPipeline::Operation) {
            if (!$currentOperationScheduled) {
              self::pushPendingOperation($pendingOperations, $currentOperationId);
              $currentOperationScheduled = TRUE;
            }
            self::pushPendingOperation($pendingOperations, $source['entity']);
          }
        }
      }
      if ($currentOperationScheduled) {
        // We need more input data, execute the dependencies first.
        continue;
      }

      // Execute the operation and retrieve its output.
      $currentOperationOutput = $currentOperation->output();
      if (!$currentOperationOutput->isValid()) {
        $this->tearDown();
        $this->failedOperationId = $currentOperationId;
        $output->setErrorMessage($currentOperationOutput->errorMessage());
        return;
      }

      // Assign the output data to the connected pipeline outputs and/or
      // other operation inputs.
      foreach ($this->pipeline->targets($currentOperationId) as $sourceOutputKey => $targets) {
        foreach ($targets as $targetEntity => $targetDataSpecs) {
          foreach ($targetDataSpecs as $targetKey => $true) {
            $dataWrapper = $currentOperationOutput->at($sourceOutputKey);

            if ($targetEntity == TfPipeline::Output) {
              $output->set($targetKey, $dataWrapper);
            }
            else { // it's an operation
              $this->operations[$targetEntity]->setInput($targetKey, $dataWrapper);
              self::pushPendingOperation($pendingOperations, $targetEntity);
            }
          }
        }
      }
    }
    $this->tearDown();
  }

  private function ensureOperationInstances() {
    if (!isset($this->operations)) {
      $this->operations = $this->pipeline->createOperations();

      foreach ($this->operations as $operationId => $operation) {
        $operation->registerOutputSchemaListener($this);
        $operation->registerOutputSchemaListener($this);
      }
    }
  }

  private function tearDown() {
    $this->operations = NULL;
  }

  private static function pushPendingOperation(&$pendingOperations, $operationId) {
    if (in_array($operationId, $pendingOperations)) {
      // No operation shall be executed twice, and this condition also helps
      // to avoid circular (infinite) additions.
      return;
    }
    array_push($pendingOperations, $operationId);
  }
}
